__author__ = 'Joe Linn'

from . import action
from .action import abstractdocument
from .response import *
from .responseset import *
import pylastica.exception.bulk


class Bulk(object):
    DELIMITER = "\n"
    UDP_DEFAULT_HOST = 'localhost'
    UDP_DEFAULT_PORT = 9700

    def __init__(self, client):
        """

        @param client:
        @type client: pylastica.client.Client
        """
        assert isinstance(client, pylastica.client.Client), "client must be an instance of Client: %r" % client
        self._client = client
        self._actions = []
        self._index = ''
        self._type = ''

    @property
    def index(self):
        """
        The name of the index to be used for this operation
        @return:
        @rtype: str
        """
        return self._index

    @index.setter
    def index(self, index):
        """
        Set the index to be used for this operation
        @param index:
        @type index: str or pylastica.index.Index
        """
        if isinstance(index, pylastica.index.Index):
            index = index.name
        self._index = str(index)

    def has_index(self):
        """

        @return:
        @rtype: bool
        """
        return self.index is not None and self.index != ''

    @property
    def doc_type(self):
        """
        Return the type name for this operation
        @return:
        @rtype: str
        """
        return self._type

    @doc_type.setter
    def doc_type(self, doc_type):
        """

        @param doc_type:
        @type doc_type: str or pylastica.doc_type.DocType
        """
        if isinstance(doc_type, pylastica.doc_type.DocType):
            self.index = doc_type.index.name
            doc_type = doc_type.name
        self._type = str(doc_type)

    def has_type(self):
        """

        @return:
        @rtype: bool
        """
        return self.doc_type is not None and self.doc_type != ''

    @property
    def path(self):
        """

        @return:
        @rtype: str
        """
        path = '/'
        if self.has_index():
            path += self.index + '/'
            if self.has_type():
                path += self.doc_type + '/'
        path += '_bulk'
        return path

    def add_action(self, action):
        """
        Add a bulk action
        @param action:
        @type action: pylastica.bulk.action.Action
        @return:
        @rtype: self
        """
        assert isinstance(action, pylastica.bulk.action.Action), "action must be an instance of Action: %r" % action
        self._actions.append(action)
        return self

    def add_actions(self, actions):
        """
        Add multiple actions
        @param actions:
        @type actions: list of pylastica.bulk.action.Action
        @return:
        @rtype: self
        """
        for action in actions:
            self.add_action(action)
        return self

    @property
    def actions(self):
        """

        @return:
        @rtype: list of pylastica.bulk.action.Action
        """
        return self._actions

    def add_document(self, document, op_type=None):
        """

        @param document:
        @type document: pylastica.document.Document
        @param op_type: bulk operation type
        @type op_type: str
        @return:
        @rtype: self
        """
        from pylastica.bulk.action import AbstractDocument
        return self.add_action(AbstractDocument.create(document, op_type))

    def add_documents(self, documents, op_type=None):
        """

        @param documents:
        @type documents: list of pylastica.document.Document
        @param op_type: bulk operation type
        @type op_type: str
        @return:
        @rtype: self
        """
        for doc in documents:
            self.add_document(doc, op_type)
        return self

    def add_raw_data(self, data):
        """
        Add raw bulk data according to the ES bulk protocol
        @param data:
        @type data: list of dict
        @return:
        @rtype: self
        """
        bulk_action = None
        for row in data:
            if isinstance(row, dict):
                op_type = row.keys()[0]
                metadata = row[op_type]
                if pylastica.bulk.action.Action.is_valid_op_type(op_type):
                    #add previous action
                    if bulk_action is not None:
                        self.add_action(bulk_action)
                    bulk_action = pylastica.bulk.action.Action(op_type, metadata)
                elif isinstance(bulk_action, pylastica.bulk.action.Action):
                    bulk_action.source = row
                    self.add_action(bulk_action)
                    bulk_action = None
                else:
                    raise pylastica.exception.InvalidException("Invalid bulk data. Source must follow action metadata.")
            else:
                raise pylastica.exception.InvalidException("Invalid bulk data. Should be list of dict, Document, or Bulk.Action")
        #add last action if available
        if bulk_action is not None:
            self.add_action(bulk_action)
        return self

    def add_script(self, script, op_type=None):
        """
        Add a script
        @param script: Script object
        @type script: pylastica.script.Script
        @param op_type: bulk operation
        @type op_type: str
        @return:
        @rtype: self
        """
        action = abstractdocument.AbstractDocument.create(script, op_type)
        return self.add_action(action)

    def add_scripts(self, scripts, op_type=None):
        """
        Add multiple scripts
        @param scripts:
        @type scripts: list of pylastica.script.Script
        @param op_type: bulk operation
        @type op_type: str
        @return:
        @rtype: self
        """
        for script in scripts:
            self.add_script(script)
        return self

    def add_data(self, data, op_type=None):
        """
        Add data (document or script)
        @param data:
        @type data: pylastica.document.Document or pylastica.script.Script or list of pylastica.document.Document or list of pylastica.script.Script
        @param op_type: bulk operation
        @type op_type: str
        @return:
        @rtype: self
        """
        if not isinstance(data, list):
            data = [data]
        for action_data in data:
            if isinstance(action_data, pylastica.script.Script):
                self.add_script(action_data)
            elif isinstance(action_data, pylastica.document.Document):
                self.add_document(action_data)
            else:
                raise TypeError("Data must be a Document, a Script, or a list comprised of either or both of those types: %r" % action_data)
        return self

    def to_string(self):
        """

        @return:
        @rtype: str
        """
        data = ''
        for action in self.actions:
            data += str(action)
        return data

    def __str__(self):
        """

        @return:
        @rtype: str
        """
        return self.to_string()

    def to_list(self):
        """

        @return:
        @rtype: list
        """
        data = []
        for action in self.actions:
            for row in action.to_list():
                data.append(row)
        return data

    def send(self):
        """
        Send this bulk action to the server
        @return:
        @rtype: pylastica.bulk.responseset.ResponseSet
        """
        return self._process_response(self._client.request(self.path, pylastica.request.Request.PUT, str(self)))

    def _process_response(self, response):
        """

        @param response:
        @type response: pylastica.response.Response
        @return:
        @rtype: pylastica.bulk.responseset.ResponseSet
        """
        from pylastica.bulk.action import AbstractDocument
        response_data = response.data
        actions = self.actions
        bulk_responses = []
        if 'items' in response_data and isinstance(response_data['items'], list):
            for key, item in enumerate(response_data['items']):
                try:
                    action = actions[key]
                except ValueError:
                    raise pylastica.exception.InvalidException("No response found for action #%s." % key)
                op_type = action.op_type
                bulk_response_data = item[item.keys()[0]]
                if isinstance(action, AbstractDocument):
                    data = action.get_data()
                    if isinstance(data, pylastica.document.Document) and data.auto_populate or self._client.get_config_value(['document', 'autoPopulate'], False):
                        if not data.has_id() and '_id' in bulk_response_data:
                            data.doc_id = bulk_response_data['_id']
                        if '_version' in bulk_response_data:
                            data.version = bulk_response_data['_version']
                bulk_responses.append(pylastica.bulk.response.Response(bulk_response_data, action, op_type))
        bulk_response_set = pylastica.bulk.responseset.ResponseSet(response, bulk_responses)
        if bulk_response_set.has_error():
            raise pylastica.exception.bulk.ResponseException(bulk_response_set)
        return bulk_response_set

    def send_udp(self, host=None, port=None):
        """

        @param host:
        @type host: str
        @param port:
        @type port: int
        @return:
        @rtype:
        """
        if host is None:
            host = self._client.get_config_value(['udp', 'host'], self.UDP_DEFAULT_HOST)
        if port is None:
            port = self._client.get_config_value(['udp', 'port'], self.UDP_DEFAULT_PORT)
        message = str(self)
        #TODO: implement udp socket stuff
        raise NotImplementedError("Haven't implemented UDP bulk transmission yet.")



__author__ = 'Joe Linn'

#import pylastica
import pylastica.request

class Settings(object):
    DEFAULT_REFRESH_INTERVAL = '1s'

    def __init__(self, index):
        """

        @param index:
        @type index: pylastica.index.Index
        """
        assert isinstance(index, pylastica.index.Index), "index must be of type Index: %r" % index
        self._index = index
        self._response = None
        self._data = {}

    def get(self, setting=None):
        """
        Returns the current settings of the index
        @param setting: optional setting name to return
        @type setting: str
        @return:
        @rtype: dict or str or None
        """
        data = self.request().data
        settings = data[self.index.name]['settings']['index']
        if setting is not None:
            if setting in settings:
                return settings[setting]
            else:
                return None
        return settings

    def set(self, data):
        """
        Set or update index settings
        @param data: arguments
        @type data: dict
        @return:
        @rtype: pylastica.response.Response
        """
        return self.request(data, pylastica.request.Request.PUT)

    def set_number_of_replicas(self, replicas):
        """
        Set the number of replicas
        @param replicas:
        @type replicas: int
        @return:
        @rtype: pylastica.response.Response
        """
        return self.set({'number_of_replicas': int(replicas)})

    def set_read_only(self, read_only=True):
        """
        Sets the index to read only
        @param read_only:
        @type read_only: bool
        @return:
        @rtype: pylastica.response.Response
        """
        return self.set({'blocks.read_only': read_only})

    def get_blocks_read(self):
        """

        @return:
        @rtype: bool
        """
        return bool(self.get('blocks.read'))

    def set_blocks_read(self, state=True):
        """

        @param state:
        @type state: bool
        @return:
        @rtype: pylastica.response.Response
        """
        state = 1 if state else 0
        return self.set({'blocks.read': state})

    def get_blocks_write(self):
        """

        @return:
        @rtype: bool
        """
        return self.get('blocks.write')

    def set_blocks_write(self, state=True):
        """

        @param state:
        @type state: bool
        @return:
        @rtype: pylastica.response.Response
        """
        state = 1 if state else 0
        return self.set({'blocks.write': int(state)})

    def get_blocks_metadata(self):
        """

        @return:
        @rtype: bool
        """
        return bool(self.get('blocks.metadata'))

    def set_blocks_metadata(self, state=True):
        """

        @param state:
        @type state: bool
        @return:
        @rtype: pylastica.response.Response
        """
        state = 1 if state else 0
        return self.set({'blocks.metadata': int(state)})

    def set_refresh_interval(self, interval):
        """
        Set the index refresh interval
        @param interval: 3s for 3 seconds, 5m for 5 minutes, etc. -1 to disable
        @type interval: str
        @return:
        @rtype: pylastica.response.Response
        """
        return self.set({'refresh_interval': interval})

    def get_refresh_interval(self):
        """
        Get the refresh interval
        @return:
        @rtype: str
        """
        interval = self.get('refresh_interval')
        if interval is None or interval == '':
            interval = self.DEFAULT_REFRESH_INTERVAL
        return interval

    def get_merge_policy_type(self):
        """
        Return merge policy
        @return:
        @rtype: str
        """
        return self.get('merge.policy.type')

    def set_merge_policy_type(self, type):
        """
        Set merge policy

        @param type:
        @type type: str
        @return:
        @rtype: pylastica.response.Response
        """
        self.index.close()
        response = self.set({'merge.policy.type': type})
        self.index.open()
        return response

    def set_merge_policy(self, key, value):
        """
        Set specific merge policies
        @param key: merge policy key (expunge_deletes_allowed, for example)
        @type key: str
        @param value:
        @type value: str
        @return:
        @rtype: pylastica.response.Response
        """
        self.index.close()
        response = self.set({'merge.policy.%s' % key: value})
        self.index.open()
        return response

    def get_merge_policy(self, key):
        """
        Returns a specific merge policy value
        @param key: merge policy key
        @type key: str
        @return:
        @rtype: str
        """
        return self.get('merge.policy.%s' % key)

    @property
    def index(self):
        """
        Return the index object
        @return:
        @rtype: pylastica.index.Index
        """
        return self._index

    def request(self, data=None, method=pylastica.request.Request.GET):
        """
        Update the given settings for the index
        @param data: optional data dictionary
        @type data: dict
        @param method: REST method
        @type method: str
        @return:
        @rtype: pylastica.response.Response
        """
        data = {'index': data}
        return self.index.request('_settings', method, data)
